import re as _re
import urllib as _urllib

import bs4 as _bs4
import PIL as _PIL

from . import _internal_utils as _utils
from . import cache as _cache

__all__ = ["Teacher"]

class Teacher:
    """This class represents a FEUP teacher as seen from their sigarra webpage.

    Please note that some attributes may be set to None, depending on whether or not
    that attribute was able to be parsed. For example: personal_webpage, email, and
    presentation are not always available in teacher pages.

    Properties:
        p_codigo         (int) # e.g. 230756
        name             (str) # e.g. 'João António Correia Lopes'
        acronym          (str) # e.g. 'JCL'
        status           (str) # e.g. 'Active'
        links            (tuple of strings) # e.g. ('http://www.fe.up.pt/~jlopes/', 'https://www.authenticus.pt/R-000-6RX', 'http://orcid.org/0000-0002-9040-0889')
        personal_webpage (str) # e.g. 'http://www.fe.up.pt/~jlopes/'
        url              (str) # e.g. 'https://sigarra.up.pt/feup/en/func_geral.formview?p_codigo=230756'
        voip             (int) # e.g. 3375
        email            (str) # e.g. None (This particular attribute could not be parsed from the page)
        rooms            (str) # e.g. 'I129'
        category         (str) # e.g. 'Professor Auxiliar'
        career           (str) # e.g. 'Pessoal Docente de Universidades'
        profession       (str) # e.g. 'Docente'
        department       (str) # e.g. 'Department of Informatics Engineering'
        presentation     (str) # e.g. 'Personal Presentation\\nJoão Correia Lopes is an Assistant Professor in Informatics Engineering...'

    Methods:
        from_url   (class method)
        from_a_tag (class method)
        picture
    
    Operators:
        __repr__, __str__
        __eq__, __le__, __lt__, __ge__, __gt__ (Comparisons between teachers and hashing are made with the p_codigo)
        __hash__
    """
    __slots__ = ["p_codigo", "name", "acronym", "status", "links", "personal_webpage", "url", "voip",
                 "email", "rooms", "category", "career", "profession", "department", "presentation"]

    def __init__(self, p_codigo : int, use_cache : bool = True):
        """Parses the webpage of the teacher with the given p_codigo.
        The cache can be bypassed by setting use_cache to False.
        If a given attribute can't be parsed (or is nonexistent),
        it will be set to None.
        """

        for attribute in self.__slots__:
            setattr(self, attribute, None)

        self.p_codigo = p_codigo
        self.url = _utils.SIG_URLS["teacher"] + "?" + _urllib.parse.urlencode({"p_codigo" : str(p_codigo)})
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "O funcionário indicado não foi encontrado." in html:
            raise ValueError(f"Teacher with p_codigo {p_codigo} doesn't exist")
        
        personal_info = soup.find("div", {"class" : "informacao-pessoal-dados-dados"})

        for row in personal_info.find_all("tr"):
            if "Name:" in str(row):
                self.name = row.find_all("td")[1].b.string

                if row.find("a") != None:
                    self.personal_webpage = row.a["href"]
            
            elif "Acronym:" in str(row):
                self.acronym = row.find_all("td")[1].string
            
            elif "Status:" in str(row):
                self.status = row.find_all("td")[1].string

            elif "Institutional E-mail:" in str(row):
                self.email = row.a.contents[0] + "@" + row.a.contents[-1]
                
            elif "Voip:" in str(row):
                self.voip = int(row.find_all("td")[1].string)

            elif "Rooms:" in str(row):
                self.rooms = row.find_all("td")[1].a.string

        self.links = tuple(tag["href"] for tag in soup.find("table", {"class" : "tabelasz"}).find_all("a"))

        functions_div = soup.find("div", {"class" : "informacao-pessoal-funcoes"})

        for td in functions_div.find_all("td", {"class" : "topo"}):

            if "Department:" in str(td):

                for row in td.find_all("tr"):
                    if "Category:" in str(row):
                        self.category = row.find_all("td")[1].string
                    
                    elif "Career:" in str(row):
                        self.career = row.find_all("td")[1].string

                    elif "Professional Group:" in str(row):
                        self.profession = row.find_all("td")[1].string
                    
                    elif "Department:" in str(row):
                        self.department = row.find_all("td")[1].string.strip()
                
                break
            
        
        personal_presentation = soup.find("div", {"class" : "informacao-pessoal-apresentacao"})

        if personal_presentation != None:
            self.presentation = personal_presentation.text.strip()
    
    def picture(self) -> _PIL.Image:
        """Returns a picture of the teacher as a PIL.Image object"""
        return _utils.get_image(_utils.SIG_URLS["picture"], {"pct_cod" : str(self.p_codigo)})

    @classmethod
    def from_url(cls, url : str, use_cache : bool = True):
        """Scrapes the teacher webpage from the given url and returns a Teacher object"""
        
        matches = _re.findall(r"p_codigo=(\d+)$", url)
        
        if len(matches) == 0:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid teacher url")
        
        p_codigo = int(matches[0])

        return Teacher(p_codigo, use_cache)
    
    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True):
        """Scrapes the teacher webpage from the given anchor tag and returns a Teacher object"""
        
        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return Teacher.from_url(bs4_tag["href"], use_cache)
    
    
    # Comparisons between teachers are made with the p_codigo
    def __eq__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo == other.p_codigo
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo > other.p_codigo
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo >= other.p_codigo
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo < other.p_codigo
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo <= other.p_codigo
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self.p_codigo)
    
    def __repr__(self):
        return f"Teacher({self.p_codigo})"
    
    def __str__(self):
        return f"{self.name} ({self.acronym})"



import getpass as _getpass
import os as _os
import re as _re
import urllib as _urllib

import requests as _requests
import requests_futures.sessions as _sessions

from . import _internal_utils as _utils

__all__ = ["Credentials"]

class Credentials:
    """This class represents your login credentials. It also
    features a non-persistent cache that stores htmls of urls
    previously accessed.

    Properties:
        username (int) # e.g. 201806185
        session  (requests.Session object) # The session that has the login cookies
        cache    (dict) # Maps a url string to an html string

    Methods:
        get_html
        get_html_async
        download
    
    Operators:
        __repr__
    """
    __slots__ = ["username", "session", "cache"]

    def __init__(self, username : int = None, password : str = None):
        """Creates a Credentials object that can be used to access pages with priviledged access.

        Credentials(123456789, "password1") #This will obviously fail
        ValueError: The login credencials aren't valid

        For example, if you happen to be a student, you have privileged access to the course units you are taking.
        If the function doesn't receive a password and/or a username, it will ask the user for them.
        Note: username must be an int and password must a string.
        """

        #This is a modified version of the https://github.com/msramalho/sigpy login function

        AUTH_FAIL = "O conjunto utilizador/senha não é válido." # If this string appears on the HTML answer, the login creds aren't valid:(

        if username == None:
            username = int(input("Username?\n:> up"))
        if password == None:
            password = _getpass.getpass(f"Password for {username}?\n:> ")
        
        if type(username) != int: # Making sure the arguments are strings
            raise TypeError(f"login() 'username' argument must be an int, not '{type(username).__name__}'")
        if type(password) != str:
            raise TypeError(f"login() 'password' argument must be a string, not '{type(password).__name__}'")

        session = _requests.Session()
        credencials = {'p_user': username, 'p_pass': password}
        request = session.post(_utils.SIG_URLS["authentication"], params = credencials) # Sending the creds to sigarra
        if AUTH_FAIL in request.text:
            raise ValueError("The login credencials aren't valid")
        
        self.username = username
        self.session  = session
        self.cache    = {}
    
    def get_html(self, url : str, params : dict = {}):
        """Functionally equivalent to requests.get(url, params).text,
        with scripts and styles removed. The result is stored in cache.
        """
        if params != {}:
            url = url + "?" + _urllib.parse.urlencode(params, doseq = True) # Emulating the RequestsIII library
        
        if url in self.cache:
            pass
        else: # We need to get the page from the web
            request = self.session.get(url) # Getting the webpage
            request.raise_for_status()  # If the request fails, I want to know about it
            html = request.text

            self.cache[url] = _utils.trim_html(html)
        
        return self.cache[url]

    def get_html_async(self, urls, n_workers : int = 10):
        """get_html, but async, give or take.
        Takes a list (or any iterable) of urls and returns a corresponding generator of htmls.
        The htmls have their scripts and styles removed and are stored in cache.
        """

        urls = tuple(urls)

        work_queue = (url for url in urls if url not in self.cache)

        with _sessions.FuturesSession(max_workers = n_workers) as session:
            session.cookies.update(self.session.cookies) # Getting the cookies from the object session to this assynchronous session

            futures = [session.get(url) for url in work_queue]

            for future in futures:
                request = future.result()
                print("hey")

                if request.status_code != 200:
                    continue # GTFO
                
                url  = request.url
                html = request.text

                self.cache[url] = _utils.trim_html(html)
        
        return (self.get_html(url) for url in urls)
    
    def download(self, url : str, folder_path : str) -> str:
        """Downloads the file from the given url and saves it
        to the given folder path. It returns the file path.
        If the path doesn't exist, it will be created automatically.
        Note: it only works with sigarra."""

        # https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
        # https://stackoverflow.com/questions/31804799/how-to-get-pdf-filename-with-python-requests

        if "conteudos_service.conteudos_cont" not in url:
            raise ValueError(f"the url argument {url} is not a valid sigarra download url")
        
        # create the cache folder if it does not exist
        if not _os.path.exists(folder_path):
            _os.makedirs(folder_path)
        
        with self.session.get(url) as r:
            r.raise_for_status()
            filename = _re.findall("filename=(.+)", r.headers['content-disposition'])[0][1:-1]
            
            path = _os.path.join(folder_path, filename)
            with open(path, 'wb') as f:
                for chunk in r.iter_content(chunk_size= 2 ** 15):
                    if chunk: # filter out keep-alive new chunks
                        f.write(chunk)
        
        return path
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}
    
    def __repr__(self):
        return f"Credentials({self.username})"



import datetime as _datetime
import re as _re
import urllib as _urllib

import bs4 as _bs4

from . import _Credentials
from . import cache as _cache
from . import exams as _exams
from . import _Student
from . import _Teacher
from . import timetable as _timetable
from . import _internal_utils as _utils

__all__ = ["CurricularUnit"]

class CurricularUnit:
    """This class represents a FEUP curricular unit.

    Properties:
        pv_ocorrencia_id   (int)
        url                (str) # The url of this curricular unit page
        name               (str) # e.g. 'Microprocessors and Personal Computers'
        code               (str) # e.g. 'EIC0016'
        acronym            (str) # e.g. 'MPCP'
        academic_year      (int) # e.g. 2018 (meaning this curricular unit was taught in 2018/2019)
        semester           (int) # either 1 or 2
        has_moodle         (bool)
        is_active          (bool)
        webpage_url        (str or None) # For example, CurricularUnit(419983).webpage_url == 'https://web.fe.up.pt/~jlopes/doku.php/teach/fpro/index' (FPRO)
        number_of_students (int)
        curricular_year    (int) # usually 1-5
        ECTS_credits       (float)
        regents            (tuple of Teacher objects)
        teachers           (tuple of Teacher objects)
        text               (string) # Basically a text dump starting from "Teaching language"
    
    Methods:
        from_url   (class method)
        from_a_tag (class method)
        contents
        students
        timetable
        all_timetables
        other_occurrences
        stats
        grades_distribution
        statistics_history
        classes
        exams
        results
    
    Operators:
        __repr__, __str__
        __eq__, __le__, __lt__, __ge__, __gt__ (Comparisons between curricular units and hashing are made with the pv_ocorrencia_id)
        __hash__
    """
    __slots__ = ["pv_ocorrencia_id", "url", "name", "code", "acronym", "academic_year", "semester", "has_moodle", "is_active", "webpage_url", 
                 "number_of_students", "curricular_year", "ECTS_credits", "regents", "teachers", "text"]
    
    def __init__(self, pv_ocorrencia_id : int, use_cache : bool = True):
        """Parses the webpage of the curricular unit with the given pv_ocorrencia_id.
        The cache can be bypassed by setting use_cache to False.
        If a given attribute can't be parsed (or is nonexistent),
        it will be set to None.
        """

        self.pv_ocorrencia_id = pv_ocorrencia_id
        self.url = _utils.SIG_URLS["curricular unit"] + "?" + _urllib.parse.urlencode({"pv_ocorrencia_id" : str(pv_ocorrencia_id)})
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "The School responsible for the occurrence was not found." in html:
            raise ValueError(f"Curricular unit with pv_ocorrencia_id {pv_ocorrencia_id} doesn't exist")
       
        contents = soup.find("div", {"id" : "conteudoinner"})

        self.name = contents.find_all("h1")[1].string

        first_table = contents.find("table")
        self.code, self.acronym = _utils.scrape_html_table(first_table)[0][1], _utils.scrape_html_table(first_table)[0][4]

        self.academic_year = _utils.parse_academic_year(contents)

        match = _re.findall(r"Instance: \d\d\d\d/\d\d\d\d - (\d)S", str(contents))[0]
        self.semester = int(match)

        self.has_moodle = "UC tem página no Moodle" in str(contents)

        second_table = first_table.find_next("table", {"class" : "formulario"})
        self.webpage_url = None
        for label, info in _utils.scrape_html_table(second_table):
            if   label == "Active? ":           # Usually the first row on the table
                self.is_active = "Yes" in info
            elif label == "Web Page: ":         # Usually the second row on the table (optional)
                self.webpage_url = info
        
        third_table = second_table.find_next("table", {"class" : "dados"})
        third_table = _utils.scrape_html_table(third_table)
        third_table.pop(0) # We don't care about the header row

        n_students = (int(row[1]) for row in third_table)
        self.number_of_students = sum(n_students)

        self.curricular_year = int(third_table[0][3])
        self.ECTS_credits    = float(third_table[0][5].replace(',', '.'))
        
        teachers_div   = contents.find("div", {"class" : "horas"})
        if teachers_div == None:
            teachers_links = filter(lambda tag: "func_geral.formview" in str(tag), contents.find_all("a"))
        else:
            teachers_links = filter(lambda tag: "func_geral.formview" in tag["href"], teachers_div.find_all("a"))
        teachers_urls = [_utils.BASE_URL + tag["href"] for tag in teachers_links]
        _cache.get_html_async(teachers_urls)
        self.teachers  = tuple(_Teacher.Teacher.from_url(url) for url in teachers_urls)

        regents_div    = contents.find("div", {"class" : "responsabilidades"})
        if regents_div == None:
            self.regents = ()
        else:
            regents_links  = filter(lambda tag: "p_codigo" in tag["href"], regents_div.find_all("a")) # If "p_codigo" is in the url, then it's a teacher
            self.regents   = tuple((_Teacher.Teacher.from_a_tag(link) for link in regents_links))

        text = contents.text
        beggining_index = text.find("Teaching language")
        self.text = text[beggining_index:]

    def contents(self, credentials : _Credentials.Credentials) -> dict:
        """Returns a nested dictionary structure, where each dictionary represents a folder.
        Every dictionary maps a string (the folder or file name) to either a dictionary
        (a nested folder) or a tuple (a file or a link).

        A tuple representing a file is made of 4 attributes:
        (
            file type       :string (either "file" or "link"),
            url             :string (note: if you want to download a file from sigarra, check Credentials.download()),
            info            :string or None,
            upload date     :Datetime.Date (Note: links will have this attribute always set to today)
        )

        Example of the output:
        {'A folder': {'folderception': {'A file': ('file',
                                                'https://sigarra.up.pt/feup/pt/conteudos_service.conteudos_cont?pct_id=012345&pv_cod=06ahastCa',
                                                None,
                                                datetime.date(2018, 10, 2)),
                                        'Python': ('file',
                                                'https://sigarra.up.pt/feup/pt/conteudos_service.conteudos_cont?pct_id=012346&pv_cod=06TtaSaPH7',
                                                'aaaaaa',
                                                datetime.date(2018, 11, 13)),
                                        'Simulator': ('link',
                                                    'https://github.com/hneemann/Digital',
                                                    'It\'s a simulator.',
                                                    datetime.date(2019, 5, 25))},
                    'jhbh': {'aa': ('link',
                                    'https://salmanarif.bitbucket.io/visual/index.html',
                                    'aaaaaa',
                                    datetime.date(2019, 8, 2)),
                            'bb': ('file',
                                    'https://sigarra.up.pt/feup/pt/conteudos_service.conteudos_cont?pct_id=012347&pv_cod=06WjyvGato',
                                    'banana',
                                    datetime.date(2018, 10, 12))}}}
        """
        html = credentials.get_html(self.url) # curricular unit page
        soup = _bs4.BeautifulSoup(html, "lxml")

        tag = soup.find("a", {"title" : "Contacts"})

        if tag == None:
            return {}
            # The curricular unit either doesn't have a contents page
            # or you don't have access to it
            # I can't distinguish between these two situations, therefore I'm
            # returning an empty dict. I prefer this approach to raising an (perhaps erroneous) exception
        
        url = _utils.BASE_URL_PT + tag["href"]

        html = credentials.get_html(url) # contents page
        soup = _bs4.BeautifulSoup(html, "lxml")
        content = soup.find("div", {"id" : "conteudoinner"})

        if "Não existem conteúdos para ver" in html:
            return {}
        
        def has_pct_grupo(a):
            try:
                int(a["name"])
                return True
            except:
                return False

        pct_grupos = (a["name"] for a in content.find_all("a") if has_pct_grupo(a))
        query = "&".join(f"pct_grupo={grupo}" for grupo in pct_grupos)
        url = url + "&" + query 

        html = credentials.get_html(url) # contents page with all the folders open
        soup = _bs4.BeautifulSoup(html, "lxml")
        content = soup.find("div", {"id" : "conteudoinner"})
        
        files_and_dirs = [p for p in content.find_all("p") if p.has_attr("class")]
        
        def convert_list_to_dict(files_and_dirs_list : list, result : dict, conteudosnivel : int = 1):
            """Please use protective glasses when looking at this function"""
            for i, p in enumerate(files_and_dirs_list):
                if p["class"] == [f"conteudosnivel{conteudosnivel}"]:
                    if "/feup/pt/imagens/Pasta" in str(p): # It's a directory
                        name = p.find_all("a")[-1].string
                        result[name] = {}

                        for i1, p1 in enumerate(files_and_dirs_list):
                            if i1 <= i:
                                continue

                            if p1["class"] == [f"conteudosnivel{conteudosnivel}"]:
                                convert_list_to_dict(files_and_dirs_list[i + 1: i1], result[name], conteudosnivel + 1)
                                break
                        else:
                            convert_list_to_dict(files_and_dirs_list[i + 1:], result[name], conteudosnivel + 1)
                    
                    else: # It's a file (or a link)
                        if p.a == None: # If there's no link, just ignore it
                            continue

                        file_type = "link" if p.find("span", {"class" : "t"}) == None else "file"
                        name = p.a.string
                        try:
                            info = p.find("span", {"class" : "textopequenoconteudos"}).string
                        except:
                            info = None

                        if file_type == "file":
                            
                            upload_time = _datetime.date(*map(int, p.find("span", {"class" : "t"}).string.split("-")[1].split("/")))
                            url = _utils.BASE_URL_PT + p.a["href"]
                        else:
                            upload_time = _datetime.date.today() # A placeholder value
                            url = p.a["href"]

                        result[name] = (file_type, url, info, upload_time)

        result = {}
        convert_list_to_dict(files_and_dirs, result, 1)

        return result

    def students(self, credentials : _Credentials.Credentials, use_cache : bool = True) -> list:
        """Parses the student page and returns a list of tuples.
        A credencials object is required because one needs 
        privileged access to access the page.
        The cache can be bypassed by setting use_cache to False.
        The each tuple is made of a Student object, their status (string), 
        number of registrations (int), and the type of student.
        Example of a tuple:
        [
            ...,
            (Student(201812345), 'Ordinário', 1, 'Normal'),
            ...
        ]
        """

        data   = []
        tables = []

        def format_table(tags_list, index): # Note: what this function returns doesn't matter,
            if index == 0:                  # we're appending the information in the tables to data
                return None                               
            
            student_username = int(tags_list[0].string)
            student = _Student.Student(student_username)

            status = tags_list[2].string.strip()

            number_of_registrations = int(tags_list[3].string)

            student_type = tags_list[4].string

            data.append((student, status, number_of_registrations, student_type))
                    
        n_pages    = (self.number_of_students // 50) + 1
        pages_urls = []
        for n in range(1, n_pages + 1):
            url = _utils.SIG_URLS["curricular unit students"] + "?" + _urllib.parse.urlencode({"pv_ocorrencia_id" : str(self.pv_ocorrencia_id), "pv_num_pag" : str(n)})
            pages_urls.append(url)
        
        for html in credentials.get_html_async(pages_urls):
            soup = _bs4.BeautifulSoup(html, "lxml")
            table = soup.find("table", {"class" : "dadossz"})
            tables.append(table)


        student_urls = []
        for table in tables: # get all the student urls
            student_usernames = _re.findall(r"(\d\d\d\d\d\d\d\d\d)", str(table))
            student_urls.extend(_utils.SIG_URLS["student page"] + "?" + _urllib.parse.urlencode({"pv_num_unico" : username}) for username in student_usernames)
        _cache.get_html_async(student_urls, use_cache = use_cache) # Refreshing the cache

        for table in tables:
            _utils.scrape_html_table(table, format_table) # get the data in the tables
        
        return data
    
    def timetable(self, credentials : _Credentials.Credentials) -> list:
        """Returns the current curricular unit timetable 
        as a list of dictionaries if possible, otherwise returns None.
        (see timetable.parse_current_timetable for more info)
        Example:
        [
            {'class type': 'T',
            'classes': ('1MIEIC01',
                        '1MIEIC02',
                        '1MIEIC09'),
            'curricular unit': CurricularUnit(419989),
            'finish': datetime.time(11, 0),
            'room': ('B003',),
            'start': datetime.time(10, 0),
            'teachers': (Teacher(23545),),
            'weekday': 'Friday'},
            ...
        ]
        """
        html = credentials.get_html(_utils.SIG_URLS["curricular unit timetable"], {"pv_ocorrencia_id" : self.pv_ocorrencia_id})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_current_timetable(credentials, soup.a["href"])
    
    def all_timetables(self, credentials : _Credentials.Credentials) -> dict:
        """Parses all the timetables related to this curricular unit.
        Returns a dictionary which maps a tuple with two datetime.date objects,
        start and finish (the time span in which this timetable is valid), to a
        list of dictionaries (see timetable.parse_timetable for further info).
        (see timetable.parse_timetables for further info)
        An example:
        {
            (datetime.date(2019, 2, 10), datetime.date(2019, 6, 1)): [
                {...},
                {...},
                {...},
                ...
            ],
            ...
        }
        """
        html = credentials.get_html(_utils.SIG_URLS["curricular unit timetable"], {"pv_ocorrencia_id" : self.pv_ocorrencia_id})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_timetables(credentials, soup.a["href"])

    def other_occurrences(self, use_cache : bool = True) -> tuple:
        """Returns a tuple of CurricularUnit objects,
        which are occurrences of this object from other years.
        The cache can be bypassed by setting use_cache to False.
        """
        html = _cache.get_html(self.url)
        soup = _bs4.BeautifulSoup(html, "lxml")

        tag = soup.find("a", {"title" : "Other occurrences"})
        url = _utils.BASE_URL + tag["href"]

        html = _cache.get_html(url)         # Now we have the html of the "other occurrences" page 
        soup = _bs4.BeautifulSoup(html, "lxml")

        table = soup.find_all("table", {"class" : "dados"})[1]
        course_tags = table.find_all("a")

        _cache.get_html_async((_utils.BASE_URL + a_tag["href"] for a_tag in course_tags), use_cache = use_cache) # Refresh the cache

        return tuple(CurricularUnit.from_a_tag(a_tag) for a_tag in course_tags)
        
    def stats(self, credentials : _Credentials.Credentials) -> tuple:
        """Returns a tuple of 3 ints: number of registered students,
        number of evaluated students, and number of approved students.
        A Credentials object is required.
        """
        
        html = credentials.get_html(_utils.SIG_URLS["curricular unit statistics"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não foram encontrados estudantes inscritos na ocorrência indicada." in html:
            raise LookupError(f"The statistics for {self.__repr__()} have not been released yet")

        table = soup.find_all("table")[-1] # get the last table of the page
        table_data = table.find_all("td", {"class" : "k n"})[0:3] # get the first three tds

        registered, evaluated, approved = map(lambda tag: int(tag.string), table_data)

        return (registered, evaluated, approved)
    
    def grades_distribution(self, credentials : _Credentials.Credentials) -> dict:
        """Returns a dictionary representing the distribution of the grades.
        The dictionary maps the grade to the number of students who got
        that grade. Here is an example of the output of this function:
        {'RFC': 9,
         'RFE': 17,
         'RFF': 18,
             5: 4,
             6: 5,
             7: 7,
             8: 6,
             9: 4,
            10: 12,
            11: 11,
            12: 13,
            13: 21,
            14: 4,
            15: 12,
            16: 7,
            17: 13,
            18: 8,
            19: 7,
            20: 1}
        """

        html = credentials.get_html(_utils.SIG_URLS["curricular unit grades distribution"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não foram encontrados estudantes avaliados na ocorrência indicada." in html:
            raise LookupError(f"The statistics for {self.__repr__()} have not been released yet")
        
        table = soup.find_all("table")[-2] # get the second last table of the page
        tbody = table.find("tbody")
        result = {}

        def format_table(tags_list, index): # Note: what this function returns doesn't matter,
            try:                            # We're updating the result dictionary
                grade = int(tags_list[0].string)
            except:
                grade = tags_list[0].string
            
            n_students = int(tags_list[2].string)

            result[grade] = n_students
        
        _utils.scrape_html_table(tbody, format_table)

        return result

    def statistics_history(self, credentials : _Credentials.Credentials) -> list:
        """Returns a list of tuples each representing an academic year.
        Contents of a tuple:
        (
            academic year        (int)
            registered students, (int)
            evaluated students,  (int)
            approved students,   (int)
            evaluated students average grade,      (float)
            evaluated students standard deviation, (float)
            approved students  average grade,      (float)
            approved students  standard deviation, (float)
        )

        Example of the output of this function:
       [(2016, 208, 152, 120, 12.43, 3.75, 13.4,  2.5),
        (2017, 203, 149, 113, 12.08, 3.08, 12.27, 3.0),
        (2018, 203, 170, 133, 12.51, 3.76, 13.53, 2.76)]
        """
        
        html = credentials.get_html(_utils.SIG_URLS["curricular unit stats history"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id), "pv_n_prev_alet" : "20"})
        soup = _bs4.BeautifulSoup(html, "lxml")

        table = soup.find_all("table")[-1] # get the last table of the page
        tbody = table.find("tbody")
        rows = [row for row in tbody.find_all("tr") if row["class"] == ['p']]

        result = []
        
        for row in rows:
            academic_year_td = row.find("td", {"class" : "k"})
            academic_year = int(academic_year_td.text[0:4])

            row_data = row.find_all("td", {"class" : "n"})

            registered, evaluated, approved = map(lambda tag: int(tag.string), row_data[0:3])

            evaluated_avg, evaluated_std_dev, approved_avg, approved_std_dev = map(lambda tag: float(tag.string), row_data[3:7])

            result.append((academic_year, registered, evaluated, approved, evaluated_avg, evaluated_std_dev, approved_avg, approved_std_dev))

        return result
    
    def classes(self, credentials : _Credentials.Credentials, use_cache : bool = True) -> dict:
        """Returns a dictionary which maps a class name to a tuple of students
        (the students of that class). Example:
        {
            '1MIEIC01': (Student(201800000),
                         Student(201800001),
                         Student(201800002),
                         Student(201800003)),
            '1MIEIC02': (Student(201800004),
                         Student(201800005),
                         Student(201800006),
                         Student(201800007),
                         Student(201800008),
                         Student(201800009),
                         Student(201800010),
                         Student(201800011),
                         Student(201800012)),
            '...'     : (...)
        }
        """
        html = credentials.get_html(_utils.SIG_URLS["curricular unit classes"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id)}) # this html redirects us to the url we want
        soup = _bs4.BeautifulSoup(html, "lxml")

        tags = soup.find_all("a")

        if len(tags) == 1:
            urls = [tags[0]["href"]]
        else:
            urls = [_utils.BASE_URL + tag["href"] for tag in tags if "it_listagem.lista_turma_disciplina" in str(tag)]

        result = {}

        for url in urls:
            html = credentials.get_html(url)
            soup = _bs4.BeautifulSoup(html, "lxml")
            contents = soup.find("div", {"id" : "conteudo"})

            all_students_urls = (_utils.BASE_URL + tag["href"] for tag in contents.find_all("a") if tag.parent.name == "td")
            _cache.get_html_async(all_students_urls, use_cache = use_cache) # refresh the cache
            
            title = contents.find("h3") # starting point

            while title.find_next("h3") != None:
                title = title.find_next("h3") # Skip to the next class

                class_name = _re.findall(r"Turma: ([^\s]+)", title.text)[0]

                table = title.find_next("table")

                if table == None:
                    students = ()
                else:
                    students = tuple(_Student.Student.from_a_tag(tag) for tag in table.find_all("a"))

                result[class_name] = students

        return result

    def exams(self, use_cache : bool = False):
        """Returns a tuple of dictionaries (See exams.exams() for further information)"""
        url = _utils.SIG_URLS["curricular unit exams"] + "?" + _urllib.parse.urlencode({"p_ocorr_id" : str(self.pv_ocorrencia_id)})

        return _exams.exams(url, use_cache)

    def results(self, credentials : _Credentials.Credentials, use_cache : bool = True) -> dict:
        """Returns a dictionary which maps a string representing the exams season ('Época Normal (2ºS)'/'Época Recurso (2ºS)')
        to a list of tuples with two values each. The first value is a Student object and the second value is the grade the student got.
        Example:
        {'Época Normal (2ºS)': [(Student(201800001), 10),
                                (Student(201800002), 13),
                                (Student(201800003), 10),
                                (Student(201800004), 'RFE'),
                                (Student(201800005), 'RFF'),
                                (Student(201800006), 'RFF'),
                                ...],
        'Época Recurso (2ºS)': [(Student(201800008), 11),
                                (Student(201800009), 7),
                                (Student(201800010), 8),
                                (Student(201800011), 8),
                                (Student(201800012), 'RFE'),
                                (Student(201800013), 13),
                                (Student(201800014), 5),
                                (Student(201800019), 'RFC'),
                                ...]}
        """
        html = credentials.get_html(_utils.SIG_URLS["curricular unit results"], params = {"pv_ocorr_id" : str(self.pv_ocorrencia_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não tem permissões para aceder a este conteúdo" in html:
            raise PermissionError("Your Credentials object does not have access to this curricular unit's results")
        
        tags = (tag for tag in soup.find_all("a") if "lres_geral.show_pauta_resul" in str(tag))

        result = {}

        for tag in tags:
            result[tag.string] = []

            url = _utils.BASE_URL + tag["href"]

            html = credentials.get_html(url)
            soup = _bs4.BeautifulSoup(html, "lxml")

            table = soup.find("table", {"class" : "dadossz"})

            rows = table.find_all("tr", {"class" : "i"})

            student_urls = []
            for row in rows: # get all the student urls
                student_usernames = _re.findall(r"(\d\d\d\d\d\d\d\d\d)", str(row))
                student_urls.extend(_utils.SIG_URLS["student page"] + "?" + _urllib.parse.urlencode({"pv_num_unico" : username}) for username in student_usernames)
            _cache.get_html_async(student_urls, use_cache = use_cache) # Refreshing the cache

            for row in rows:
                username = int(row.find("td", {"class" : "l"}).string)

                grade_str = row.find("td", {"class" : "n"}).string
                try:
                    grade = int(grade_str)
                except:
                    grade = grade_str
                
                result[tag.string].append((_Student.Student(username), grade))
        
        return result

    @classmethod
    def from_url(cls, url : str, use_cache : bool = True):
        """Scrapes the curricular unit webpage from the given url and returns a CurricularUnit object"""
        
        matches = _re.findall(r"pv_ocorrencia_id=(\d+)$", url)
        
        if len(matches) == 0:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid curricular unit url")
        
        pv_ocorrencia_id = int(matches[0])

        return CurricularUnit(pv_ocorrencia_id, use_cache)
    
    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True):
        """Scrapes the curricular unit webpage from the given anchor tag and returns a CurricularUnit object"""
        
        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return CurricularUnit.from_url(bs4_tag["href"], use_cache)
    
    # Comparisons between curricular units are made with the pv_ocorrencia_id
    def __eq__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id == other.pv_ocorrencia_id
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id > other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id >= other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id < other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id <= other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self.pv_ocorrencia_id)
    
    def __repr__(self):
        return f"CurricularUnit({self.pv_ocorrencia_id})"
    
    def __str__(self):
        return f"{self.acronym} ({self.academic_year}/{self.academic_year + 1})" # e.g. ALGE (2019/2019)



import re as _re
import urllib as _urllib

import bs4 as _bs4

from . import _Credentials
from . import _internal_utils as _utils
from . import cache as _cache
from . import _CurricularUnit
from . import exams as _exams
from . import _Teacher

__all__ = ["Course"]

class Course:
    """This class represents a course as seen from its sigarra webpage.

    Properties:
        pv_curso_id      (int) # e.g. 742
        pv_ano_lectivo   (int) # e.g. 2018
        url              (str) # e.g. 'https://sigarra.up.pt/feup/en/cur_geral.cur_view?pv_curso_id=742&pv_ano_lectivo=2018'
        name             (str) # e.g. 'Master in Informatics and Computing Engineering'
        official_code    (int) # e.g. 9459
        directors        (tuple of Teacher objects) # e.g. (Teacher(210006), Teacher(209566))
        acronym          (str) # e.g. 'MIEIC'
        text             (str or None)

    Methods:
        from_url   (class method)
        from_a_tag (class method)
        classes
        curricular_units
        syllabus
    
    Operators:
        __repr__, __str__
        __eq__, __le__, __lt__, __ge__, __gt__ (Courses are sorted firstly by pv_curso_id and secondly by year)
        __hash__
    """
    __slots__ = ["pv_curso_id","pv_ano_lectivo", "url", "name", "official_code", "directors", "acronym", "text"]

    def __init__(self, pv_curso_id : int, pv_ano_lectivo : int = None, use_cache : bool = True):
        """Parses the webpage of the course with the given pv_curso_id.
        If a pv_ano_lectivo (e.g. 2019) is not used, the current academic year will be used.
        The cache can be bypassed by setting use_cache to False.
        """
        
        if pv_ano_lectivo == None:
            pv_ano_lectivo = _utils.get_current_academic_year()

        self.pv_curso_id    = pv_curso_id
        self.pv_ano_lectivo = pv_ano_lectivo

        payload = {
            "pv_curso_id"    : str(pv_curso_id),
            "pv_ano_lectivo" : str(pv_ano_lectivo)
        }
        self.url = _utils.SIG_URLS["course"] + "?" + _urllib.parse.urlencode(payload)
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Course/Cycle of Studies nonexistent." in html or "The Organic Unit is not involved in teaching the course/CS." in html:
            raise ValueError(f"Course with pv_curso_id {pv_curso_id} and pv_ano_lectivo {pv_ano_lectivo} doesn't exist")
       
        contents = soup.find("div", {"id" : "conteudoinner"})

        self.name = contents.find_all("h1")[1].string

        info_box = contents.find("div", {"class" : "caixa-informativa"})

        directors_ids = _re.findall(r"vld_entidades_geral.entidade_pagina\?pct_codigo=(\d+)", html)
        self.directors  = tuple(_Teacher.Teacher(int(_id)) for _id in directors_ids)

        for row in info_box.find_all("tr"):
            if "Acronym:" in str(row):
                self.acronym = row.find_all("td")[1].string
            
            elif "Official Code:" in str(row):
                self.official_code = int(row.find_all("td")[1].string)
        
        try:
            self.text = contents.find("div", {"class" : "col-md-8 col-sm-6 col-xs-12"}).text
        except:
            self.text = None


    def classes(self, credentials : _Credentials.Credentials) -> dict:
        """Returns a dictionary which maps the courses' classes
        to their corresponding timetable links.
        Example:
            {'1MIEIC01':'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000001&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC02': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000002&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC03': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000003&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC04': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000004&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC05': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000005&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC06': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000006&pv_periodos=1&pv_ano_lectivo=2018',
            ...                                                                                                                   }
        You may be interested on the functions provided by timetable.py
        """

        payload = {
            "pv_curso_id"    : str(self.pv_curso_id),
            "pv_ano_lectivo" : str(self.pv_ano_lectivo),
            "pv_periodos"    : "1" # All the classes in that year
        }

        html = credentials.get_html(_utils.SIG_URLS["course classes"], payload)
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não existem dados para este ano letivo." in html:
            raise ValueError("No classes were available to be parsed")

        classes_tags = soup.find_all("a", {"class" : "t"})

        return {tag.string : _utils.BASE_URL_PT + tag["href"] for tag in classes_tags}


    def curricular_units(self, use_cache : bool = True) -> tuple:
        """Returns a tuple of CurricularUnit objects representing
        the curricular units of the course. Curricular units without
        a link are not included.
        Example:
            (CurricularUnit(419981),
             CurricularUnit(419982),
             CurricularUnit(419983),
             CurricularUnit(419984),
             CurricularUnit(419985),
             CurricularUnit(419986),
             CurricularUnit(419987),
             CurricularUnit(419988),
             CurricularUnit(419989),
             CurricularUnit(419990),
             CurricularUnit(419991),
             CurricularUnit(419992),
             CurricularUnit(419993),
             CurricularUnit(419994),
             CurricularUnit(419995),
             CurricularUnit(419996),
             ...                   )
        """
        html = _cache.get_html(self.url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        for tag in soup.find_all("a"):
            if "cur_geral.cur_planos_estudos_view" in str(tag):
                url = _utils.BASE_URL + tag["href"]
                break
        else:
            raise ValueError("No study plan link was found on the course page")
        
        html = _cache.get_html(url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        curricular_units_tags = (tag for tag in soup.find_all("a") if "ucurr_geral.ficha_uc_view" in str(tag))
        curricular_units_urls = [_utils.BASE_URL + tag["href"] for tag in curricular_units_tags]

        # The following block of code assynchronously loads both the curricular units' and the teachers' webpages
        teachers_urls = []
        for uc_html in _cache.get_html_async(curricular_units_urls, use_cache = use_cache):
            uc_soup = _bs4.BeautifulSoup(uc_html, "lxml")

            for tag in uc_soup.find_all("a"):
                if "func_geral.formview" in str(tag):
                    teachers_urls.append(_utils.BASE_URL + tag["href"])
        _cache.get_html_async(teachers_urls, use_cache = use_cache)

        return tuple(set(_CurricularUnit.CurricularUnit.from_url(url) for url in curricular_units_urls))

    
    def syllabus(self, use_cache : bool = True) -> list:
        """Returns a list of tuples.
        Example of a tuple:
        (
            'Software Engineering', # keyword (str)
            None,                   # branch (either str or None, depending on whether or not the curricular unit belongs to a specific branch)
            CurricularUnit(420015), # the curricular unit (CurricularUnit)
            True                    # whether or not it is mandatory (bool)
        )

        Example of the output:
        [
            ('Automation', 'Automation', CurricularUnit(418771), False),
            ('Automation', 'Automation', CurricularUnit(418770), False),
            ('Automation', 'Automation', CurricularUnit(418737), True),
            ('Automation', 'Automation', CurricularUnit(418746), True),
            ('Automation', 'Automation', CurricularUnit(418747), False),
            ('Automation', 'Automation', CurricularUnit(418736), True),
            ('Automation', 'Thermal Energy', CurricularUnit(418773), False),
            ('Automation', 'Thermal Energy', CurricularUnit(418736), False),
            ('Automation', 'Production Management', CurricularUnit(418773), False),
            ('Automation', 'Production Management', CurricularUnit(418736), False),
            ('Automation', 'Production, Conception and Manufacturing', CurricularUnit(418773), False),
            ('Automation', 'Structural Engineering and Machine Design', CurricularUnit(418736), False),
            ('Automation', None, CurricularUnit(418722), True),
            ('Personal and Interpersonal Skills', 'Automation', CurricularUnit(418787), False),
            ...
        ]
        """

        html = _cache.get_html(self.url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        for tag in soup.find_all("a"):
            if "cur_geral.cur_planos_estudos_view" in str(tag):
                url = _utils.BASE_URL + tag["href"]
                break
        else:
            raise ValueError("No study plan link was found on the course page")
        
        html = _cache.get_html(url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        code_to_uc = {uc.code : uc for uc in self.curricular_units(use_cache = use_cache)}
        result = []

        tables = soup.find_all("table", {"class" : "dados"})

        for table in tables:
            for row in table.find_all("tr")[1:]: # Ignore the header row
                
                keyword, branch, code, acronym, name, _credits, mandatory = (tag.string for tag in row.find_all("td"))

                try:
                    result.append((keyword, branch, code_to_uc[code], "Yes" in str(mandatory)))
                except KeyError: # The curricular unit doesn't have a page
                    continue
        
        return result

    def exams(self, use_cache : bool = False):
        """Returns a tuple of dictionaries (See exams.exams() for further information)"""
        url = _utils.SIG_URLS["curricular unit exams"] + "?" + _urllib.parse.urlencode({"p_curso_id" : str(self.pv_curso_id)})

        return _exams.exams(url, use_cache)


    @classmethod
    def from_url(cls, url : str, use_cache : bool = True):
        """Scrapes the course webpage from the given url and returns a Course object"""
        
        try:
            pv_curso_id    = int(_re.findall(r"pv_curso_id=(\d+)"   , url)[0])
            pv_ano_lectivo = int(_re.findall(r"pv_ano_lectivo=(\d+)", url)[0])
        except:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid course url")

        return Course(pv_curso_id, pv_ano_lectivo, use_cache)


    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True):
        """Scrapes the course webpage from the given anchor tag and returns a Course object"""
        
        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return Course.from_url(bs4_tag["href"], use_cache)

    
    # Comparisons between curricular units are made with the pv_curso_id and pv_ano_lectivo
    def _unique_value(self):
        return self.pv_curso_id * 10000 + self.pv_ano_lectivo

    def __eq__(self, other):
        if isinstance(other, Course):
            return self._unique_value() == other._unique_value()
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, Course):
            return self._unique_value() > other._unique_value()
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, Course):
            return self._unique_value() >= other._unique_value()
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, Course):
            return self._unique_value() < other._unique_value()
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, Course):
            return self._unique_value() <= other._unique_value()
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self._unique_value())
    
    def __repr__(self):
        return f"Course({self.pv_curso_id}, {self.pv_ano_lectivo})"
    
    def __str__(self):
        return f"{self.acronym} ({self.pv_ano_lectivo}/{self.pv_ano_lectivo + 1})"



# -*- coding: utf-8 -*-

"""Top-level package for feupy."""

__author__ = """Daniel Filipe Amaro Monteiro"""
__email__ = 'up201806185@fe.up.pt'
__version__ = '0.1.1'

from . _Student import Student
from . _Teacher import Teacher
from . _CurricularUnit import CurricularUnit
from . _Course import Course

from . _Credentials import Credentials
from . _User import User

from . import exams
from . import timetable


import datetime as _datetime

import bs4 as _bs4

from . import _internal_utils as _utils
from . import cache as _cache
from . import _CurricularUnit

__all__ = ["exams"]

def exams(url : str, use_cache : bool = False) -> tuple:
    """Returns a tuple made of dictionaries.
    Example of a dictionary:

        {'curricular unit': CurricularUnit(420037),
         'finish': datetime.datetime(2019, 9, 6, 20, 0),
         'observations': 'Tenho um exame de outra disciplina marcado para esse horário e assim juntava os dois.\\r\\nJosé Luís Moura Borges\\r\\n',
         'rooms': None, # A tuple of strings, if available
         'season': 'Especial de Conclusão - SET-E-ESPECIAL',
         'start': datetime.datetime(2019, 9, 6, 17, 0)}
     
    If a value is not available (like 'rooms', in this case), it will be set to none.
    Important note: use_cache = True will only apply to the exams pages, not the curricular units in those exams pages.
    """
    html = _cache.get_html(url, use_cache = use_cache)
    soup = _bs4.BeautifulSoup(html, 'lxml')
    content = soup.find("div", {"id" : "conteudoinner"})

    tags = [tag for tag in content.find_all("a")]
    tags.pop(0)

    urls = [_utils.BASE_URL_PT + tag["href"] for tag in tags if "exa_geral.exame_view" in tag["href"]]
    _cache.get_html_async(urls, use_cache = use_cache) # Refresh the cache
    return tuple(_parse_exam_page(url) for url in urls)

def _parse_exam_page(url : str):
    """Parses an exam page from the url and returns a dictionary"""

    html = _cache.get_html(url)
    soup = _bs4.BeautifulSoup(html, 'lxml')
    content = soup.find("div", {"id" : "conteudoinner"})
    table = content.find("table")

    rows = table.find_all("tr")

    code, _, season, date, start, length = rows[:6]

    curricular_unit = _CurricularUnit.CurricularUnit.from_a_tag(code.a)

    season = season.find_all("td")[1].string

    year, month, day = map(int, date.find_all("td")[1].string.split("-"))

    start_hour, start_minute = map(int, start.find_all("td")[1].string.split(":"))

    start = _datetime.datetime(year, month, day, start_hour, start_minute)

    length_hour, length_minute = map(int, length.find_all("td")[1].string.split(":"))

    delta = _datetime.timedelta(hours = length_hour, minutes = length_minute)

    finish = start + delta

    rooms = observations = None
    for row in rows[6:]:
        if "Salas:" in str(row):
            rooms = tuple(tag.string for tag in row.find_all("a"))
        elif "Observações:" in str(row):
            observations = row.find_all("td")[1].string
    
    return {
        "curricular unit" : curricular_unit,
        "season"          : season,
        "start"           : start,
        "finish"          : finish,
        "rooms"           : rooms,
        "observations"    : observations
    }



import re as _re

import bs4 as _bs4

from . import _Credentials
from . import _internal_utils as _utils
from . import _Course
from . import _CurricularUnit
from . import timetable as _timetable

__all__ = ["User"]

class User:
    """This class represents the information that can be extracted
    from your personal webpage.

    Properties:
        pv_fest_id      (int) # A sort of course specific student identification number.
                              # See from_credentials() and get_pv_fest_ids() for ways to get
                              # this number
        course          (Course object) # The course related to this pv_fest_id
        credentials     (Credentials object) # The credentials argument from the __init__ function
                                             # (This is done to avoid having to pass the same
                                             # Credentials object to every function)

    Methods:
        from_credentials (class  method)
        get_pv_fest_ids  (static method)
        courses_units
        timetable
        all_timetables
        classes
    """

    __slots__ = ["pv_fest_id", "course", "credentials"]

    def __init__(self, pv_fest_id : int, credentials: _Credentials.Credentials):
        """Parses your personal webpage. See User.get_pv_fest_ids() and User.from_credentials()
        for ways to get your pv_fest_id. A pv_fest_id is like a "course specific student id",
        from what I can tell.
        """
        self.pv_fest_id = pv_fest_id
        self.credentials = credentials # Note, this is only a reference

        html = credentials.get_html(_utils.SIG_URLS["courses units"], {"pv_fest_id" : str(pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")
        
        if "Não tem permissões para aceder a este conteúdo." in html:
            raise ValueError("Your Credentials object does not have permission to access the page related to this pv_fest_id")

        for tag in soup.find_all("h2"):
            if "cur_geral.cur_view" in str(tag):
                self.course = _Course.Course.from_a_tag(tag.a)
                break
        else:
            raise Exception("No course was found")

    def courses_units(self) -> list:
        """Returns a list of tuples. Each tuple contains a Curricular unit which represents
        a curricular unit that you have had in a previous year or you currently enrolled in,
        and either an int which represents the grade you got at that curricular unit or None
        if a grade is not available.

        Example:
            [(CurricularUnit(419981), 10),
             (CurricularUnit(419982), 11),
             (CurricularUnit(419983), 12),
             (CurricularUnit(419984), 13),
             (CurricularUnit(419985), 14),
             (CurricularUnit(420521), 15),
             (CurricularUnit(419986), None),
             (CurricularUnit(419987), None),
             (CurricularUnit(419988), None),
             (CurricularUnit(419989), None),
             (CurricularUnit(419990), None)]
        """
        html = self.credentials.get_html(_utils.SIG_URLS["courses units"], {"pv_fest_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        result = []
        for row in soup.find_all("tr", {"class" : "d"}):
            curricular_unit = _CurricularUnit.CurricularUnit.from_a_tag(row.a)
            
            try:
                grade = int(row.find("td", {"class" : "n"}).string)
            except:
                grade = None
            
            result.append((curricular_unit, grade))
        
        return result

    def timetable(self) -> list:
        """Returns the current user timetable 
        as a list of dictionaries if possible, otherwise returns None.
        (see timetable.parse_current_timetable for more info)
        Example:
        [
            {'class type': 'T',
            'classes': ('1MIEIC01',
                        '1MIEIC02',
                        '1MIEIC09'),
            'curricular unit': CurricularUnit(419989),
            'finish': datetime.time(11, 0),
            'room': ('B003',),
            'start': datetime.time(10, 0),
            'teachers': (Teacher(23545),),
            'weekday': 'Friday'},
            ...
        ]
        """
        html = self.credentials.get_html(_utils.SIG_URLS["personal timetable"], {"pv_fest_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_current_timetable(self.credentials, soup.a["href"])
    
    def all_timetables(self) -> dict:
        """Parses all the timetables related to this user.
        Returns a dictionary which maps a tuple with two datetime.date objects,
        start and finish (the time span in which this timetable is valid), to a
        list of dictionaries (see timetable.parse_timetable for further info).
        (see timetable.parse_timetables for further info)
        An example:
        {
            (datetime.date(2019, 2, 10), datetime.date(2019, 6, 1)): [
                {...},
                {...},
                {...},
                ...
            ],
            ...
        }
        """
        html = self.credentials.get_html(_utils.SIG_URLS["personal timetable"], {"pv_fest_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_timetables(self.credentials, soup.a["href"])

    def classes(self):
        """Returns a list of tuples, each tuple containing a CurricularUnit object
        and the class name as a string.
        """
        raise NotImplementedError("No idea if this works or not")
        html = self.credentials.get_html(_utils.SIG_URLS["classes data"] , params = {"pv_estudante_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, 'lxml')
    
        tables = soup.find_all("table", {"class" : "tabela"})[1:] # Forget first table

        result = []
        for table in tables:
            for row in table.find_all("tr")[2:]: #Forget the header rows
                curricular_unit = _CurricularUnit.CurricularUnit.from_a_tag(row.find("a"))
                class_name = row.find_all("td")[3].string

                result.append((curricular_unit, class_name))
        
        return result

    @classmethod
    def from_credentials(cls, credentials: _Credentials.Credentials):
        """Returns a User object made from the first result from get_pv_fest_ids().
        Usually, students are enrolled in only one course, which means that
        get_pv_fest_ids() tends to be a tuple with a single int.
        """
        return User(User.get_pv_fest_ids(credentials)[0], credentials)


    @staticmethod
    def get_pv_fest_ids(credentials: _Credentials.Credentials) -> tuple:
        """Returns a tuple of ints, each representing a pv_fest_id."""
        payload = {"pv_num_unico" : str(credentials.username)}
        html = credentials.get_html(_utils.SIG_URLS["student page"], payload)
    
        matches = _re.findall(r"pv_fest_id=(\d+)", html)

        if len(matches) == 0:
            raise ValueError("No pv_fest_id's were found")
    
        return tuple(int(match) for match in matches)

    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}



import atexit as _atexit
import datetime as _datetime
import os as _os
import random as _random
import re as _re
import shelve as _shelve
import time as _time
import urllib as _urllib

import bs4 as _bs4
import requests as _requests
import requests_futures.sessions as _sessions

from . import _internal_utils as _utils

__all__ = ["load_cache", "get_html", "reset", "remove_invalid_entries", "get_html_async"]

cache = None # This will eventually be the cache object
# The cache is a dictionary-like object whose values are structured in the following way:
# {
#     url0 : (timeout0, html0),
#     url1 : (timeout1, html1),
#     ...
# }
# In which url is a string, timeout is an int or a float (which represents the "due by date" as seconds since epoch), and html is a string
#
# When get_html is called, if the url is not in the cache keys (or the key is present but the cache entry is outdated),
# get_html will first check if the url matches any key in custom_treatments (it matches if key(url) returns True).
# If so the cache will call custom_treatments[key](html), instead of default_treatment(html)
#
# This behaviour was implemented so that commonly accessed pages (curricular units, students, teachers)
# can have their htmls even more shrinked than usual (by returning a timeout and the div with the useful information)

_custom_treatments = {}


############################### Custom treatments ################################

def _is_curricular_unit(url):
    return "ucurr_geral.ficha_uc_view" in url 

def _is_teacher(url):
    return "func_geral.formview" in url

def _is_student(url):
    return "fest_geral.cursos_list" in url

def _trim_curricular_unit(html):
    """Returns a heavily trimmed version of the curricular unit html"""
    html = _utils.trim_html(html)

    soup = _bs4.BeautifulSoup(html, "lxml")

    useful_stuff = soup.find("div", {"id" : "envolvente"})

    useful_stuff.find("div", {"id" : "colunaprincipal"}).decompose() # Remove the left side-bar

    return str(useful_stuff)

def _trim_teacher(html):
    """Returns a heavily trimmed version of the teacher html"""
    html = _utils.trim_html(html)

    soup = _bs4.BeautifulSoup(html, "lxml")

    useful_stuff = soup.find("div", {"id" : "conteudo"})

    return str(useful_stuff)

def _trim_student(html):
    """Returns a heavily trimmed version of the student html"""
    html = _utils.trim_html(html)

    soup = _bs4.BeautifulSoup(html, "lxml")

    useful_stuff = soup.find("div", {"id" : "conteudo-extra"})

    return str(useful_stuff)

def _curricular_unit_treatment(html):
    uc_academic_year = _utils.parse_academic_year(html)

    if uc_academic_year < _utils.get_current_academic_year():
        timeout = _time.time() + _random_radioactive_lifetime(_datetime.timedelta(days = 6*30)) # It is extremely unlikely that a uc page from a previous year will ever change
        return (timeout, _trim_curricular_unit(html))
    else:
        timeout = _time.time() + _random_radioactive_lifetime(_datetime.timedelta(days = 30))
        return (timeout, _trim_curricular_unit(html))

def _teacher_treatment(html):
    timeout = _time.time() + _random_radioactive_lifetime(_datetime.timedelta(days = 2*30))
    return (timeout, _trim_teacher(html))

def _student_treatment(html):
    year  = _utils.get_current_academic_year() + 1
    timeout = _datetime.datetime(year, 9, 20).timestamp() # Only update at the beginning of the next academic year
    return (timeout, _trim_student(html))

############################ End of custom treatments ############################

_custom_treatments = {
    _is_curricular_unit : _curricular_unit_treatment,
    _is_teacher         : _teacher_treatment,
    _is_student         : _student_treatment
}

def load_cache(flag = "c", path = None):
    """Loads the cache from disk and stores it in the variable cache. 
    If the variable cache is different than None, the function will do nothing.

    Refer to this to know what the flag parameter does:
    https://docs.python.org/3/library/dbm.html#dbm.open

    If you wish to save the cache to a folder of your chooosing, 
    set path to an existing directory path.
    """
    global cache

    if cache != None: # There is something already in cache
        return
    
    if path == None:
        package_root = _os.path.join(_os.path.dirname(__file__), "..")

        path = _os.path.join(package_root, "data")
    
    cache = _shelve.open(_os.path.join(path, "cache"), flag = flag)
    
    _atexit.register(cache.close) # This function will be called at program termination, in order to make sure that the cache is saved to disk
    # the cache object itself would most likely save itself on exit, but better be safe than sorry

def _random_radioactive_lifetime(half_life = _datetime.timedelta(days=2), cutoff = True):
    """Returns a randomly generated lifetime of a radioactive particle with a half-life of half_life seconds
    (a time_delta is also acceptable).
    If cutoff is True (which is by default), if the lifetime generated is bigger than half_life * 4, the function
    will instead return half_life * 4, which is useful if you wish to avoid ridiculously large lifetimes
    This function is used to disperse the timeouts of the cache values.
    """
    if isinstance(half_life, _datetime.timedelta):
        half_life = half_life.total_seconds()

    lambd = 0.693147 / half_life # ln(2)/half-life
    lifetime = _random.expovariate(lambd)

    return min(lifetime, half_life * 4) if cutoff else lifetime

def _default_treatment(html):
    """The default treatment for any url that doesn't match any key in custom_treatments.
    Returns a tuple made of a timeout provided by random_radioactive_lifetime and the html
    without any scripts nor styles, to save space"""

    return (_time.time() + _random_radioactive_lifetime(), _utils.trim_html(html))

def _cache_entry_is_valid(key):
    """Returns True if the cache entry is valid, otherwise returns false"""
    global cache
    return cache[key][0] > _time.time()

def get_html(url, params = {}, use_cache = True):
    """More or less functionally equivalent to requests.get(url, params).text, with the added
    benefit of a persistent cache with customizable html treatment and timeouts, depending on the url.
    
    The curricular units' pages, along with the students' and teachers' pages,
    are modified to reduce their memory footprint. The default treatment removes scripts
    and styles from the html to also reduce their size. If you wish to bypass the cache, set
    use_cache to False, (please note that the html modifications will still be made either way).

    Helpful note: if you know that you are going to make a crapton of requests beforehand, you probably
    should call get_html_async first to populate the cache.
    """
    global cache

    if cache == None:
        load_cache()
    
    if params != {}:
        url = url + "?" + _urllib.parse.urlencode(params, doseq = True) # Emulating the RequestsIII library
    
    if url in cache and _cache_entry_is_valid(url) and use_cache:
        pass
    else: # We need to get the page from the web
        request = _requests.get(url) # Getting the webpage
        request.raise_for_status()  # If the request fails, I want to know about it
        html = request.text

        for match_rule, custom_treatment in _custom_treatments.items(): # Can we apply any custom treatment to the html?
            if match_rule(url):
                cache[url] = custom_treatment(html)
                break
        else:                                                         # If not, use the default treatment
            cache[url] = _default_treatment(html)
    
    return cache[url][1]

def reset():
    """Eliminates all entries from this cache implementation"""
    global cache

    if cache == None:
        load_cache()

    cache.clear()

def remove_invalid_entries(urls = None):
    """Removes all the cache entries in urls that have timed out"""
    global cache

    if cache == None:
        load_cache()

    if urls == None:
        urls = cache.keys()
    
    hit_list = set() # Each element must be unique, otherwise it will be popped twice

    for url in urls:
        if url in cache and not _cache_entry_is_valid(url):
            hit_list.add(url)
    
    for url in hit_list:
        cache.pop(url)

def get_html_async(urls, n_workers = 10, use_cache = True):
    """get_html, but async, give or take.
    Takes a list (or any iterable) of urls and returns a corresponding generator of htmls.
    The htmls have their scripts and styles removed and are stored in cache.
    """
    global cache

    if cache == None:
        load_cache()

    urls = tuple(urls)

    remove_invalid_entries(urls)

    work_queue = (url for url in urls if url not in cache or not use_cache)

    with _sessions.FuturesSession(max_workers = n_workers) as session:
        
        futures = [session.get(url) for url in work_queue]

        for future in futures:
            request = future.result()

            if request.status_code != 200:
                continue # GTFO
            
            url  = request.url
            html = request.text

            for match_rule, custom_treatment in _custom_treatments.items(): # Can we apply any custom treatment to the html?
                if match_rule(url):
                    cache[url] = custom_treatment(html)
                    break
            else:                                                         # If not, use the default treatment
                cache[url] = _default_treatment(html)
    
    return (get_html(url) for url in urls)
